#include "curve.h"
#include "xmem.h"

static real_coordinate_type int_to_real_coord (coordinate_type);

curve_type
new_curve (void)
{
  curve_type curve;
  XMALLOC (curve, sizeof (struct curve));
  curve->point_list = NULL;
  CURVE_LENGTH (curve) = 0;
  CURVE_CYCLIC (curve) = false;
  CURVE_START_TANGENT (curve) = CURVE_END_TANGENT (curve) = NULL;
  PREVIOUS_CURVE (curve) = NEXT_CURVE (curve) = NULL;

  return curve;
}

curve_type
copy_most_of_curve (curve_type old_curve)
{
  curve_type curve = new_curve ();

  CURVE_CYCLIC (curve) = CURVE_CYCLIC (old_curve);
  PREVIOUS_CURVE (curve) = PREVIOUS_CURVE (old_curve);
  NEXT_CURVE (curve) = NEXT_CURVE (old_curve);

  return curve;
}

void
free_curve (curve_type curve)
{
  if (CURVE_LENGTH (curve) > 0)
    free (curve->point_list);
  if (CURVE_START_TANGENT (curve))
    free (CURVE_START_TANGENT (curve));
  if (CURVE_END_TANGENT (curve))
    free (CURVE_END_TANGENT (curve));
}


void
append_pixel (curve_type curve, coordinate_type coord)
{
  append_point (curve, int_to_real_coord (coord));
}


void
append_point (curve_type curve, real_coordinate_type coord)
{
  CURVE_LENGTH (curve)++;
  XREALLOC (curve->point_list, CURVE_LENGTH (curve) * sizeof(point_type));
  LAST_CURVE_POINT (curve) = coord;
}


#define NUM_TO_PRINT 3


curve_list_type
new_curve_list (void)
{
  curve_list_type curve_list;

  curve_list.length = 0;
  curve_list.data = NULL;

  return curve_list;
}


void
free_curve_list (curve_list_type *curve_list)
{
  unsigned this_curve;

  for (this_curve = 0; this_curve < curve_list->length; this_curve++)
    {
      free_curve (curve_list->data[this_curve]);
      free (curve_list->data[this_curve]);
    }

  if (curve_list->data != NULL)
    free (curve_list->data);
}

void
append_curve (curve_list_type *curve_list, curve_type curve)
{
  curve_list->length++;
  XREALLOC (curve_list->data, curve_list->length * sizeof (curve_type));
  curve_list->data[curve_list->length - 1] = curve; 
}

curve_list_array_type
new_curve_list_array (void)
{
  curve_list_array_type curve_list_array;

  CURVE_LIST_ARRAY_LENGTH (curve_list_array) = 0;
  curve_list_array.data = NULL;

  return curve_list_array;
}


void
free_curve_list_array (curve_list_array_type *curve_list_array)
{
  unsigned this_list;

  for (this_list = 0; this_list < CURVE_LIST_ARRAY_LENGTH (*curve_list_array);
       this_list++) {
    free_curve_list (&CURVE_LIST_ARRAY_ELT (*curve_list_array, this_list));
  }


  if (curve_list_array->data != NULL)
    free (curve_list_array->data);
}


void
append_curve_list (curve_list_array_type *curve_list_array, curve_list_type curve_list)
{
  CURVE_LIST_ARRAY_LENGTH (*curve_list_array)++;
  XREALLOC (curve_list_array->data, CURVE_LIST_ARRAY_LENGTH (*curve_list_array) * sizeof (curve_list_type));
  LAST_CURVE_LIST_ARRAY_ELT (*curve_list_array) = curve_list;
}


static real_coordinate_type
int_to_real_coord (coordinate_type int_coord)
{
  real_coordinate_type real_coord;

  real_coord.x = int_coord.x;
  real_coord.y = int_coord.y;

  return real_coord;
}
