#ifndef CURVE_H
#define CURVE_H

#include "types.h"
#include "vector.h"


typedef struct
{
  real_coordinate_type coord;
  real t;
} point_type;


struct curve
{
  point_type *point_list;
  unsigned length;
  bool cyclic;
  vector_type *start_tangent;
  vector_type *end_tangent;
  struct curve *previous;
  struct curve *next;
};

typedef struct curve *curve_type;

#define CURVE_POINT(c, n) ((c)->point_list[n].coord)
#define LAST_CURVE_POINT(c) ((c)->point_list[(c)->length-1].coord)
#define CURVE_T(c, n) ((c)->point_list[n].t)
#define LAST_CURVE_T(c) ((c)->point_list[(c)->length-1].t)
#define CURVE_LENGTH(c)  ((c)->length)
#define CURVE_CYCLIC(c)  ((c)->cyclic)
#define CURVE_NEXT(c, n)						\
  ((n) + 1 >= CURVE_LENGTH (c)						\
  ? CURVE_CYCLIC (c) ? ((n) + 1) % CURVE_LENGTH (c) : CURVE_LENGTH (c)	\
  : (n) + 1)
#define CURVE_PREV(c, n)						\
  ((signed int) (n) - 1 < 0							\
  ? CURVE_CYCLIC (c) ? (signed int) CURVE_LENGTH (c) + (signed int) (n) - 1 : -1\
  : (signed int) (n) - 1)

#define CURVE_START_TANGENT(c) ((c)->start_tangent)
#define CURVE_END_TANGENT(c) ((c)->end_tangent)
#define PREVIOUS_CURVE(c) ((c)->previous)
#define NEXT_CURVE(c) ((c)->next)

extern curve_type new_curve (void);
extern curve_type copy_most_of_curve (curve_type c);
extern void free_curve (curve_type c);
extern void append_pixel (curve_type c, coordinate_type p);
extern void append_point (curve_type c, real_coordinate_type p);
extern void log_curve (curve_type c, bool print_t);
extern void log_entire_curve (curve_type c);
extern void display_curve (curve_type);

typedef struct
{
  curve_type *data;
  unsigned length;
  bool clockwise;
} curve_list_type;


#define CURVE_LIST_LENGTH(c_l)  ((c_l).length)
#define CURVE_LIST_ELT(c_l, n) ((c_l).data[n])
#define LAST_CURVE_LIST_ELT(c_l) ((c_l).data[CURVE_LIST_LENGTH (c_l) - 1])
#define CURVE_LIST_CLOCKWISE(c_l) ((c_l).clockwise)


extern curve_list_type new_curve_list (void);
extern void free_curve_list (curve_list_type *);
extern void append_curve (curve_list_type *, curve_type);


typedef struct
{
  curve_list_type *data;
  unsigned length;
} curve_list_array_type;


#define CURVE_LIST_ARRAY_LENGTH CURVE_LIST_LENGTH
#define CURVE_LIST_ARRAY_ELT CURVE_LIST_ELT
#define LAST_CURVE_LIST_ARRAY_ELT LAST_CURVE_LIST_ELT

extern curve_list_array_type new_curve_list_array (void);
extern void free_curve_list_array (curve_list_array_type *);
extern void append_curve_list (curve_list_array_type *, curve_list_type);

#endif 

