#ifndef SPLINE_H
#define SPLINE_H

#include <stdio.h>
#include "types.h"
#include "color.h"


typedef enum
{
  LINEARTYPE = 1, QUADRATICTYPE = 2, CUBICTYPE = 3, PARALLELELLIPSETYPE = 4,
  ELLIPSETYPE = 5, CIRCLETYPE = 6
} polynomial_degree;


typedef struct
{
  real_coordinate_type v[4];
  polynomial_degree degree;
  real linearity;
} spline_type;

#define START_POINT(spl)	((spl).v[0])
#define CONTROL1(spl)		((spl).v[1])
#define CONTROL2(spl)		((spl).v[2])
#define END_POINT(spl)		((spl).v[3])
#define SPLINE_DEGREE(spl)	((spl).degree)
#define SPLINE_LINEARITY(spl)	((spl).linearity)


extern void print_spline (FILE *, spline_type);
extern real_coordinate_type evaluate_spline (spline_type spline, real t);

typedef struct
{
  spline_type *data;
  unsigned length;
  bool clockwise;
  color_type color;
} spline_list_type;


#define SPLINE_LIST_LENGTH(s_l) ((s_l).length)
#define SPLINE_LIST_DATA(s_l) ((s_l).data)
#define SPLINE_LIST_ELT(s_l, index) (SPLINE_LIST_DATA (s_l)[index])
#define LAST_SPLINE_LIST_ELT(s_l) \
  (SPLINE_LIST_DATA (s_l)[SPLINE_LIST_LENGTH (s_l) - 1])
#define NEXT_SPLINE_LIST_ELT(s_l, index)				\
  SPLINE_LIST_ELT (s_l, ((index) + 1) % SPLINE_LIST_LENGTH (s_l))
#define PREV_SPLINE_LIST_ELT(s_l, index)				\
  SPLINE_LIST_ELT (s_l, index == 0					\
                        ? SPLINE_LIST_LENGTH (s_l) - 1			\
                        : index - 1)


extern spline_list_type *new_spline_list (void);
extern spline_list_type *init_spline_list (spline_type);
extern void free_spline_list (spline_list_type);
extern void append_spline (spline_list_type *s_list, spline_type s);
extern void concat_spline_lists (spline_list_type *s1, spline_list_type s2);

typedef struct
{
  spline_list_type *data;
  unsigned length;
} spline_list_array_type;


#define SPLINE_LIST_ARRAY_LENGTH SPLINE_LIST_LENGTH
#define SPLINE_LIST_ARRAY_DATA SPLINE_LIST_DATA
#define SPLINE_LIST_ARRAY_ELT SPLINE_LIST_ELT
#define LAST_SPLINE_LIST_ARRAY_ELT LAST_SPLINE_LIST_ELT


extern spline_list_array_type new_spline_list_array (void);
extern void append_spline_list (spline_list_array_type *, spline_list_type);



extern void free_spline_list_array (spline_list_array_type *);


#endif 
